import mock
from datetime import datetime

# First, try to import modern xmlrpc.client and support also old one on EPEL 7/Python 2
try:
    from xmlrpc.client import Fault
except ImportError:
    from xmlrpclib import Fault

import pytest

from blockerbugs.util.bz_interface import BlockerProposal, BZInterfaceError

bz_success = {'bugs': [{'alias': [],
                        'changes': {},
                        'id': 806497,
                        'last_change_time': datetime.utcnow()}]}


class TestProposeBlocker:
    def setup_method(self, method):
        self.ref_trackers = {'blocker': 123456, 'fe': 234567}
        self.ref_proposed = 345678
        self.ref_justification = 'because I said so'
        self.ref_tracker_type = 'Blocker'
        self.ref_milestone = 'F19-alpha'
        self.ref_user = 'joe@allthebugs.com'
        self.ref_cc = 'cc@allthebugs.com'

    def test_propose_blocker(self):
        stubbz = mock.Mock()
        stubbz.update_bugs.return_value = True
        stubbz.build_update.return_value = bz_success

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_bz._do_proposal(self.ref_trackers, self.ref_proposed,
                             self.ref_justification, self.ref_cc)

        assert stubbz.build_update.calls().once()
        assert stubbz.update_bugs.calls().once()

    def test_propose_blocker_comment(self):
        stubbz = mock.Mock()
        stubbz._update_bug.return_value = True
        stubbz.build_update.return_value = bz_success

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_bz.propose_bugs(self.ref_user, self.ref_milestone,
                             self.ref_justification)
        test_comment = stubbz.build_update.mock_calls[0][2]['comment']
        assert test_comment.startswith('Proposed as a Blocker for F19-alpha by joe@allthebugs.com')

    def test_propose_blocker_cc(self):
        stubbz = mock.Mock()
        stubbz._update_bug.return_value = True
        stubbz.build_update.return_value = bz_success

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, bz_user=self.ref_cc)
        test_bz.propose_bugs(self.ref_user, self.ref_milestone,
                             self.ref_justification)

        test_cc = stubbz.build_update.mock_calls[0][2]['cc_add']

        assert self.ref_cc in test_cc

    def test_propose_multiple_tracker(self):
        stubbz = mock.Mock()
        stubbz._update_bug.return_value = True
        stubbz.build_update.return_value = bz_success

        ref_trackerids = list(self.ref_trackers.values())

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, is_fe=True)
        test_bz.propose_bugs(self.ref_user, self.ref_milestone,
                             self.ref_justification)

        test_trackers = stubbz.build_update.mock_calls[0][2]['blocks_add']

        assert test_trackers == ref_trackerids

    def test_propose_single_tracker(self):
        stubbz = mock.Mock()
        stubbz._update_bug.return_value = True
        stubbz.build_update.return_value = bz_success

        ref_trackerid = [self.ref_trackers['blocker']]

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_bz.propose_bugs(self.ref_user, self.ref_milestone,
                             self.ref_justification)

        test_trackers = stubbz.build_update.mock_calls[0][2]['blocks_add']

        assert test_trackers == ref_trackerid

    def test_proposed_cc_invalid(self):
        stubbz = mock.Mock()
        stubbz._update_bug.return_value = True
        stubbz.build_update.return_value = bz_success

        stubbz.update_bugs = mock.Mock(side_effect=Fault(50, 'not 51 exception'))

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                    is_blocker = True, bz_user = self.ref_cc)

        with pytest.raises(BZInterfaceError) as excinfo:
            test_bz.propose_bugs(self.ref_user, self.ref_milestone,
                                    self.ref_justification)

        assert excinfo.value.msg == 'not 51 exception'


class TestBugProposalTrackerCheck:
    def setup_method(self, method):
        self.ref_trackers = {'blocker': 123456, 'fe': 234567}
        self.ref_justification = 'because I said so'
        self.ref_user = 'joe@allthebugs.com'
        self.ref_tracker_type = 'Blocker'
        self.ref_milestone = 'F19-alpha'
        self.ref_proposed = 345678

    def test_check_proposal_calls(self):
        refbug = mock.MagicMock()
        refbug.blocked.return_value = []
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_bz.check_blocker_proposal()

        assert stubbz.getbug.calls().once()

    def test_check_blocker_proposal_clean(self):
        refbug = mock.MagicMock()
        refbug.blocked.return_value = []
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_result = test_bz.check_blocker_proposal()

        assert test_result

    def test_check_propose_blocker_already(self):
        refbug = mock.MagicMock()
        refbug.blocked = [self.ref_trackers['blocker']]
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        test_result = test_bz.check_blocker_proposal()

        assert test_result == False

    def test_check_bothpropose_blocker_already(self):
        self.ref_tracker_type = 'Blocker and Freeze Exception'

        refbug = mock.MagicMock()
        refbug.blocked = [self.ref_trackers['blocker']]
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, is_fe=True)
        test_result = test_bz.check_blocker_proposal()

        assert test_result == False

    def test_check_bothpropose_blocker_already_fe_ok(self):
        self.ref_tracker_type = 'Blocker and Freeze Exception'

        refbug = mock.MagicMock()
        refbug.blocked = [self.ref_trackers['blocker']]
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, is_fe=True)
        test_result = test_bz.check_fe_proposal()

        assert test_result == True

    def test_check_bothpropose_fe_already(self):
        refbug = mock.MagicMock()
        refbug.blocked = [self.ref_trackers['fe']]
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, is_fe=True)
        test_result = test_bz.check_fe_proposal()

        assert test_result == False

    def test_check_bothpropose_fe_already_blocker_ok(self):
        refbug = mock.MagicMock()
        refbug.blocked = [self.ref_trackers['fe']]
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True, is_fe=True)
        test_result = test_bz.check_blocker_proposal()

        assert test_result == True

    def test_proposed_bug_notexist(self):
        stubbz = mock.MagicMock()
        stubbz.getbug = mock.Mock(side_effect=Fault(101, 'bug does not exist'))

        test_bz = BlockerProposal(stubbz, self.ref_proposed, self.ref_trackers,
                                  is_blocker=True)
        with pytest.raises(BZInterfaceError):
            test_bz.check_blocker_proposal()


class TestProposedBug():
    def setup_method(self, method):
        self.blocker_tracker = 123456
        self.fe_tracker = 234567
        self.ref_justification = 'because I said so'
        self.ref_user = 'joe@allthebugs.com'
        self.ref_tracker_type = 'Blocker'
        self.ref_milestone = 'F19-alpha'
        self.ref_proposed = 345678

    def test_proposed_bug_closed(self):
        ref_status = 'CLOSED NOTABUG'
        refbug = mock.MagicMock()
        refbug.blocked = [self.blocker_tracker]
        refbug.bug_status = ref_status
        stubbz = mock.MagicMock()
        stubbz.getbug.return_value = refbug

        test_bz = BlockerProposal(stubbz, self.ref_proposed,
                                  [self.blocker_tracker],
                                  self.ref_tracker_type)
        with pytest.raises(BZInterfaceError) as excinfo:
            test_bz.check_proposed_bug()

        assert excinfo.value.msg == 'Bug %i is CLOSED: %s' % (
        self.ref_proposed, ref_status)

    def test_proposed_bug_invalid(self):
        stubbz = mock.MagicMock()
        stubbz.getbug = mock.Mock(side_effect=Fault(101, 'bug does not exist'))
        test_bz = BlockerProposal(stubbz, self.ref_proposed,
                                  [self.blocker_tracker],
                                  self.ref_tracker_type)
        with pytest.raises(BZInterfaceError) as excinfo:
            test_bz.check_proposed_bug()

        assert excinfo.value.msg == 'bug does not exist'

