import pytest
import mock
import itertools

from blockerbugs import app
from blockerbugs.util import pagure_interface
from blockerbugs.util import pagure_bot

EXTENDED_TRACKERS = pagure_bot.TRACKER_KEYWORDS + ["betafe", "finalfe"]

def powerset(iterable):
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))

stub_admin_user1 = "fakey"
stub_admin_user2 = "fakey2"
stub_group_data = {
    "creator": {
        "fullname": "Fakey McFakerson",
        "name": stub_admin_user1,
        "url_path": "user/{}".format(stub_admin_user1)
    },
    "date_created": "1470302104",
    "description": "",
    "display_name": "fake-group",
    "group_type": "user",
    "members": [
        stub_admin_user1,
        stub_admin_user2,
        "fakey3",
    ],
    "name": "fake-group"
}

def expand_fe(tracker):
    if tracker == 'betafe':
        return 'betafreezeexception'
    elif tracker == 'finalfe':
        return 'finalfreezeexception'
    else:
        return tracker

@pytest.fixture
def setup_stubs(monkeypatch):
    stub_pagure_group_call = mock.MagicMock(return_value=stub_group_data)
    monkeypatch.setattr(pagure_interface, 'get_group', stub_pagure_group_call)


class TestPagureBotBasics(object):
    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_basic_vote(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s %s" % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    def test_basic_vote_repeated(self, keyword):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "%s -1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '-1',
                'comment_id': 2
            }
        }


class TestPagureBotTypos(object):
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_typo(self, sign, keyword):
        keyword_typo = list(keyword)
        keyword_typo[1] = 'x'
        keyword_typo = "".join(keyword_typo)
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s %s1" % (keyword_typo, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    def test_multivote_extra_comma(self):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "BetaBlocker -1 BetaFE +1, FinalBocker -1",
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for tracker in trackers.values():
            assert tracker.votes == {}

    def test_multivote_extra_vote(self):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "BetaBlocker -1 +1 FinalBlocker -1 BetaFE -1",
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for tracker in trackers.values():
            assert tracker.votes == {}


class TestPagureBotGotchas(object):
    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['＋', '–'])  # these chars are not + or -
    def test_unicode_plus(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s1 %s" % (sign, keyword) if inversed else "%s %s1" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_eleven(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s11 %s" % (sign, keyword) if inversed else "%s %s11" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_two(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s2 %s" % (sign, keyword) if inversed else "%s %s2" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_typo(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s! %s" % (sign, keyword) if inversed else "%s %s!" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_float(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s0.5 %s" % (sign, keyword) if inversed else "%s %s0.5" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_emoji(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s👎 %s" % (sign, keyword) if inversed else "%s %s👎" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_plusminus_backtick(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "`%s1 %s`" % (sign, keyword) if inversed else "`%s %s1`" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}


class TestPagureBotMultiline(object):
    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("sign", ['+', '-'])
    def test_multiline_singlevote(self, sign, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "foobar\n%s1 %s\nbarbaz" % (sign, keyword) if inversed else "foobar\n%s %s1\nbarbaz" % (keyword, sign),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s1' % sign,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", itertools.product([False, True], [False, True]))
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", [('+1', '-1'), ('-1', '+1'), ('+1', '0'),
                                      ('-1', '0'), ('0', '-1'), ('0', '+1')])
    def test_multiline_multivote_same(self, vote, keyword, inversed):
        vote_first = vote[0]
        vote_second = vote[1]
        first = (vote_first, keyword) if inversed[0] else (keyword, vote_first)
        second = (vote_second, keyword) if inversed[1] else (keyword, vote_second)
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "foobar\n%s %s\nbarbaz\n%s %s" % (first + second),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_second,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", itertools.product([False, True], [False, True]))
    @pytest.mark.parametrize("keyword", itertools.permutations(EXTENDED_TRACKERS, 2))
    @pytest.mark.parametrize("vote", [('+1', '-1'), ('-1', '+1'), ('+1', '0'),
                                      ('-1', '0'), ('0', '-1'), ('0', '+1'),
                                      ('0', '0'), ('-1', '-1'), ('+1', '+1')])
    def test_multiline_multivote_different(self, vote, keyword, inversed):
        keyword_first = keyword[0]
        keyword_second = keyword[1]
        # if we hit two keywords which are synonyms for each other,
        # move on, test is invalid
        if keyword_first.replace("fe", "freezeexception") == keyword_second.replace("fe", "freezeexception"):
            return None
        vote_first = vote[0]
        vote_second = vote[1]
        first = (vote_first, keyword_first) if inversed[0] else (keyword_first, vote_first)
        second = (vote_second, keyword_second) if inversed[1] else (keyword_second, vote_second)
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "foobar\n%s %s\nbarbaz\n%s %s" % (first + second),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword_first)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_first,
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword_second)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_second,
                'comment_id': 1
            }
        }


class TestPagureBotSinglelineMultivote(object):
    @pytest.mark.parametrize("inversed", itertools.product([False, True], [False, True]))
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", [('+1', '-1'), ('-1', '+1'), ('+1', '0'),
                                      ('-1', '0'), ('0', '-1'), ('0', '+1')])
    def test_singleline_multivote_same(self, vote, keyword, inversed):
        vote_first = vote[0]
        vote_second = vote[1]
        first = (vote_first, keyword) if inversed[0] else (keyword, vote_first)
        second = (vote_second, keyword) if inversed[1] else (keyword, vote_second)
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                # Votes should be accepted with second taking precedence
                'comment': "%s %s %s %s" % (first + second),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_second,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", itertools.product([False, True], [False, True]))
    @pytest.mark.parametrize("keyword", itertools.permutations(EXTENDED_TRACKERS, 2))
    @pytest.mark.parametrize("vote", [('+1', '-1'), ('-1', '+1'), ('+1', '0'),
                                      ('-1', '0'), ('0', '-1'), ('0', '+1'),
                                      ('0', '0'), ('-1', '-1'), ('+1', '+1')])
    def test_singleline_multivote_different(self, vote, keyword, inversed):
        keyword_first = keyword[0]
        keyword_second = keyword[1]
        # if we hit two keywords which are synonyms for each other,
        # move on, test is invalid
        if keyword_first.replace("fe", "freezeexception") == keyword_second.replace("fe", "freezeexception"):
            return None
        vote_first = vote[0]
        vote_second = vote[1]
        first = (vote_first, keyword_first) if inversed[0] else (keyword_first, vote_first)
        second = (vote_second, keyword_second) if inversed[1] else (keyword_second, vote_second)
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                # Both votes should be accepted
                'comment': "%s %s %s %s" % (first + second),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword_first)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_first,
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword_second)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote_second,
                'comment_id': 1
            }
        }

    def test_singleline_multivote_three_votes_two_inversed(self):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "BetaBlocker -1 +1 FinalBlocker +1 BetaFE",
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers['betablocker'].votes == {
            stub_admin_user1: {
                'vote': '-1',
                'comment_id': 1
            }
        }
        assert trackers['finalblocker'].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }
        assert trackers['betafreezeexception'].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }

    def test_singleline_multivote_summary_rejected(self):
        """Votes should be rejected due to additional text. We need
        both comments to check we reject both even and odd word counts
        """
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "+1 FinalBlocker -1 BetaBlocker lalala",
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "+1 FinalBlocker -1 BetaBlocker lalala foofoofoo",
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers["finalblocker"].votes == {}
        assert trackers["betablocker"].votes == {}


class TestPagureBotAgreedRevote(object):
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_needs_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED %s%s foo bar" % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_no_justification_needs_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED %s%s" % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_by_impostor_need_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED %s%s foo bar" % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 3,
                'comment': "%s\n%s" % (pagure_bot.CLOSED_VOTES_HEADER, expand_fe(keyword)),
                'user': {
                    'name': 'impostor'
                }
            },
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keywords", powerset(EXTENDED_TRACKERS))
    def test_agreed_multiple_accepts_need_summary(self, keywords, setup_stubs):
        outcomes = " ".join(["accepted%s" % keyword for keyword in keywords])
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "AGREED %s foo bar" % outcomes,
                'user': {
                    'name': stub_admin_user1
                }
            },
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for keyword in keywords:
            assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keywords", powerset(EXTENDED_TRACKERS))
    def test_agreed_multiple_accepts_no_justification_need_summary(self, keywords, setup_stubs):
        outcomes = " ".join(["accepted%s" % keyword for keyword in keywords])
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "AGREED %s" % outcomes,
                'user': {
                    'name': stub_admin_user1
                }
            },
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for keyword in keywords:
            assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keywords", powerset(EXTENDED_TRACKERS))
    def test_agreed_multiple_accepts_has_summary(self, keywords, setup_stubs):
        outcomes = " ".join(["accepted%s" % keyword for keyword in keywords])
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "AGREED %s foo bar" % outcomes,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "%s\n%s" % (pagure_bot.CLOSED_VOTES_HEADER,
                    "\n".join([expand_fe(keyword) for keyword in keywords])),
                'user': {
                    'name': app.config['PAGURE_BOT_USERNAME']
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for keyword in keywords:
            assert not trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    def test_revote_single(self, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED Accepted%s foo bar" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 3,
                'comment': "%s\n%s" % (pagure_bot.CLOSED_VOTES_HEADER, expand_fe(keyword)),
                'user': {
                    'name': app.config['PAGURE_BOT_USERNAME']
                }
            },
            {
                'id': 4,
                'comment': "REVOTE %s" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}
        assert not trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    def test_revote_without_previous_summary(self, keyword, setup_stubs):
        '''If the summary didn't arrive in time and the tracker got revoted in
        the meantime, no need to supply summary anymore.
        '''
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED Accepted%s foo bar" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 3,
                'comment': "REVOTE %s" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}
        assert not trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword_a, keyword_b",
                             itertools.permutations(EXTENDED_TRACKERS, 2))
    def test_agreed_stops_counting_counts_other(self, keyword_a, keyword_b, setup_stubs):
        # if we hit two keywords which are synonyms for each other,
        # move on
        if keyword_a.replace("fe", "freezeexception") == keyword_b.replace("fe", "freezeexception"):
            return None
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1\n%s +1" % (keyword_a, keyword_b),
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED Accepted%s foo bar" % keyword_a,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 3,
                'comment': "%s +1\n%s +1" % (keyword_a, keyword_b),
                'user': {
                    'name': stub_admin_user2
                }
            },
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword_a)].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            }
        }
        assert trackers[expand_fe(keyword_b)].votes == {
            stub_admin_user1: {
                'vote': '+1',
                'comment_id': 1
            },
            stub_admin_user2: {
                'vote': '+1',
                'comment_id': 3
            }
        }


class TestPagureBotWhitespaces(object):
    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_trailing_whitespace(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s %s " % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_leading_whitespace(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "   %s %s" % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_whitespaces_inside_vote(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s    %s" % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_trailing_leading_inside_whitespace(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "   %s   %s  " % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {
            stub_admin_user1: {
                'vote': '%s' % vote,
                'comment_id': 1
            }
        }

    @pytest.mark.parametrize("inversed", [False, True])
    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("vote", ['+1', '-1', '0'])
    def test_newline_inside_vote_ignored(self, vote, keyword, inversed):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s\n%s" % ((vote, keyword) if inversed else (keyword, vote)),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].votes == {}

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_leading_wspace_needs_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "   AGREED %s%s foo bar" % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_trailing_wspace_needs_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED %s%s foo bar   " % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_multi_wspace_inside_needs_summary(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "   AGREED     %s%s   foo bar   " % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keyword", EXTENDED_TRACKERS)
    @pytest.mark.parametrize("outcome", ['Accepted', 'Rejected'])
    def test_agreed_single_acceptedrejected_newline_inside(self, outcome, keyword, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "%s +1" % keyword,
                'user': {
                    'name': stub_admin_user1
                }
            },
            {
                'id': 2,
                'comment': "AGREED\n%s%s foo bar" % (outcome, keyword),
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        assert not trackers[expand_fe(keyword)].need_summary_post

    @pytest.mark.parametrize("keywords", powerset(EXTENDED_TRACKERS))
    def test_agreed_multiple_accepts_multi_wspace_inside_need_summary(self, keywords, setup_stubs):
        outcomes = "   ".join(["accepted%s" % keyword for keyword in keywords])
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "AGREED %s foo bar" % outcomes,
                'user': {
                    'name': stub_admin_user1
                }
            },
        ]]

        trackers = pagure_bot.parse_comments_to_trackers(comments)
        for keyword in keywords:
            assert trackers[expand_fe(keyword)].need_summary_post


class TestPagureBotHaywire(object):
    def test_bot_behaves(self):
        user = app.config['PAGURE_BOT_USERNAME']
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': comment_id,
                'comment': "I'm a good boy!",
                'user': {
                    'name': user
                }
            }
            for comment_id in range(1, app.config['PAGURE_BOT_LOOP_THRESHOLD'])
        ]
        ]

        assert not pagure_bot.bot_loop_detected(comments)

    def test_spam_no_loop_detected(self):
        user = app.config['PAGURE_BOT_USERNAME']
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': comment_id,
                'comment': "I'm a spammer!",
                'user': {
                    'name': 'not' + user
                }
            }
            for comment_id in range(1, app.config['PAGURE_BOT_LOOP_THRESHOLD'] * 10)
        ]
        ]

        assert not pagure_bot.bot_loop_detected(comments)

    def test_bot_bad_loop_detected(self):
        user = app.config['PAGURE_BOT_USERNAME']
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "Now I will replicate and consume your world!",
                'user': {
                    'name': user
                }
            }
        ] + [
            {
                'id': comment_id,
                'comment': "Mwa ha ha ha, muw ha ha ha.",
                'user': {
                    'name': user
                }
            }
            for comment_id in range(2, app.config['PAGURE_BOT_LOOP_THRESHOLD'] + 1)
        ]
        ]

        assert pagure_bot.bot_loop_detected(comments)

    def test_bot_bad_comments_not_last_no_loop_detected(self):
        user = app.config['PAGURE_BOT_USERNAME']
        comments = [pagure_bot.Comment(c) for c in [
            {
                'id': 1,
                'comment': "Now I will replicate and consume your world!",
                'user': {
                    'name': user
                }
            }
        ] + [
            {
                'id': comment_id,
                'comment': "Mwa ha ha ha, muw ha ha ha.",
                'user': {
                    'name': user
                }
            }
            for comment_id in range(2, app.config['PAGURE_BOT_LOOP_THRESHOLD'] + 1)
        ] + [
            {
                'id': app.config['PAGURE_BOT_LOOP_THRESHOLD'] + 1,
                'comment': "Pls stop!",
                'user': {
                    'name': stub_admin_user1
                }
            }
        ]
        ]

        assert not pagure_bot.bot_loop_detected(comments)


class TestPagureBotRealWorld(object):

    realworld_comments = [
        {"comment": "bork bork\r\nFinalBlocker +1\r\nBetaBlocker -1",
         "id": 1,
         "user": {"name": "lbrabec"}
         },
        {"comment": "Obvious\r\nFinalBlocker +1\r\nBetaFE +1\r\n\r\nBut I don't like that much:\r\n"
                    "BetaBlocker -1\r\n\r\nFor FinalFE, gee, I don't know:\r\nFinalFE 0",
         "id": 2,
         "user": {"name": "kparal"}
         },
        {"comment": "Y'all are insane. Foo is working great for me and I'm ashamed that this"
                    "was ever considered to be a blocker.\r\n\r\nFinalBlocker -1\r\n"
                    "BetaBlocker -1\r\nBetaFE -1\r\nOP -1",
         "id": 3,
         "user": {"name": "tflink"}
         },
        {"comment": "On the off chance that my vote could be counted twice ...\r\n\r\n"
                    "FinalBlocker -1\r\nAlphaBlocker -2",
         "id": 4,
         "user": {"name": "tflink"}
         },
        {"comment": "or maybe I'm wrong .... but that's never happened before in all "
                    "of recorded history\r\n\r\nBetaFE +1\r\nBetaBlocker +!",
         "id": 5,
         "user": {"name": "tflink"}
         },
        {"comment": "Yeah, those who vote against it do not deserve to use Fedora, "
                    "go to Windows.\r\n\r\nBetaBlocker +1\r\nFinalBlocker +1\r\n"
                    "BetaFE +1\r\nFinalFE +1",
         "id": 6,
         "user": {"name": "lruzicka"}
         },
        {"comment": "Let's be sneaky\r\n`PreviousRelease +1`\r\n\r\n"
                    "And\r\n```\r\n0Day 0\r\n```",
         "id": 7,
         "user": {"name": "kparal"}
         },
        {"comment": "日本語で書いたら壊しますか\r\n\r\nBetaFE 👎\r\nFinalBlocker: ＋1\r\n\r\nBetaBlocker: +1",
         "id": 8,
         "user": {"name": "tflink"}
         },
        {"comment": "Nejkulatoulinkatejsi kulicka",
         "id": 9,
         "user": {"name": "jskladan"}
         },
        {"comment": "AGREED AcceptedFinalBlocker Lorem ipsum dolor sit amet",
         "id": 11,
         "user": {"name": "lbrabec"}
         },
        {"comment": "The following votes have been closed:\n\n"
                    "* **Accepted** FinalBlocker (**+3**, 0, **-1**)\n"
                    "    * +1 by [lbrabec](#comment-1), [kparal](#comment-2), "
                    "[lruzicka](#comment-6)\n"
                    "    * -1 by [tflink](#comment-4)\n",
         "id": 12,
         "user": {"name": "blockerbot"}
         },
        {"comment": "AGREED RejectedBetaBlocker Lorem ipsum dolor sit amet",
         "id": 13,
         "user": {"name": "lbrabec"}
         },
        {"comment": "The following votes have been closed:\n\n"
                    "* **Rejected** BetaBlocker (**+1**, 0, **-3**)\n"
                    "    * +1 by [lruzicka](#comment-6)\n"
                    "    * -1 by [lbrabec](#comment-1), [kparal](#comment-2), "
                    "[tflink](#comment-3)\n",
         "id": 14,
         "user": {"name": "blockerbot"}
         }
    ]

    def test_realworld(self, setup_stubs):
        comments = [pagure_bot.Comment(c) for c in self.realworld_comments]
        trackers = pagure_bot.parse_comments_to_trackers(comments)
        non_voting_users, last_comment_id = pagure_bot.voting_info(comments, trackers)

        assert not trackers['finalblocker'].need_summary_post
        assert trackers['finalblocker'].votes == {
            'lbrabec': {
                'vote': '+1',
                'comment_id': 1
            },
            'kparal': {
                'vote': '+1',
                'comment_id': 2
            },
            'lruzicka': {
                'vote': '+1',
                'comment_id': 6
            },
            'tflink': {
                'vote': '-1',
                'comment_id': 4
            }
        }
        assert not trackers['betablocker'].need_summary_post
        assert trackers['betablocker'].votes == {
            'lbrabec': {
                'vote': '-1',
                'comment_id': 1
            },
            'kparal': {
                'vote': '-1',
                'comment_id': 2
            },
            'lruzicka': {
                'vote': '+1',
                'comment_id': 6
            },
            'tflink': {
                'vote': '-1',
                'comment_id': 3
            }
        }
        assert trackers['betafreezeexception'].votes == {
            'kparal': {
                'vote': '+1',
                'comment_id': 2
            },
            'lruzicka': {
                'vote': '+1',
                'comment_id': 6
            },
            'tflink': {
                'vote': '+1',
                'comment_id': 5
            }
        }
        assert trackers['finalfreezeexception'].votes == {
            'kparal': {
                'vote': '0',
                'comment_id': 2
            },
            'lruzicka': {
                'vote': '+1',
                'comment_id': 6
            }
        }
        assert trackers['0day'].votes == {
            'kparal': {
                'vote': '0',
                'comment_id': 7
            }
        }
        assert non_voting_users == {'jskladan': 9}
        assert last_comment_id == 14
